from datasette.app import Datasette
from .create_db import create_dbs
import pytest
import duckdb
from datasette_parquet.winging_it import ProxyConnection

@pytest.fixture(scope="session")
def datasette():
    create_dbs('./fixtures')
    metadata = {
        'plugins': {
            'datasette-parquet': {
                'trove': {
                    'directory': './fixtures'
                },
                'duckdb': {
                    'file': './fixtures/fixtures.duckdb'
                }

            }
        }
    }

    return Datasette(
        [],
        memory=True,
        metadata=metadata,
    )

@pytest.mark.asyncio
async def test_plugin_is_installed(datasette):
    response = await datasette.client.get("/-/plugins.json")
    assert response.status_code == 200
    installed_plugins = {p["name"] for p in response.json()}
    assert "datasette-parquet" in installed_plugins

@pytest.mark.asyncio
async def test_file_mode(datasette):
    response = await datasette.client.get('/duckdb')
    assert response.status_code == 200

@pytest.mark.asyncio
async def test_directory_mode(datasette):
    response = await datasette.client.get('/trove')
    assert response.status_code == 200

@pytest.mark.asyncio
async def test_json_works(datasette):
    response = await datasette.client.get("/trove/fixtures.json?_size=max&_labels=on&_shape=objects")
    assert response.status_code == 200
    assert response.json()['rows'] == [{'date': '2023-01-01', 'ts': '2023-01-02T03:04:05' }]

@pytest.mark.asyncio
async def test_extraneous_parameters(datasette):
    response = await datasette.client.get("/trove?sql=select+%2A+from+fixtures&_hide_sql=1")
    assert response.status_code == 200

@pytest.mark.asyncio
async def test_sql_json(datasette):
    response = await datasette.client.get("/trove.json?sql=select+%2A+from+fixtures&_hide_sql=1")
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_duckdb_table(datasette):
    response = await datasette.client.get("/duckdb/fixtures")
    assert response.status_code == 200

def test_fetchone():
    raw_conn = duckdb.connect()
    conn = ProxyConnection(raw_conn)
    fetched = conn.execute('SELECT 1 AS col').fetchone()
    assert fetched['col'] == 1



from datetime import date
from .winging_it import Row

def monkey_patch():
    from datasette.utils import CustomJSONEncoder

    original_default = CustomJSONEncoder.default

    def patched_default(self, obj):
        # We serialize dates as their ISO string.
        # An alternative: we could serialize them with metadata, eg
        # { 'type': 'date', 'value': '2022-12-25' }
        #
        # ...but I think metadata should be a separate thing.
        if isinstance(obj, date):
            return obj.isoformat()
        if isinstance(obj, Row):
            return tuple(obj)
        return original_default(self, obj)

    CustomJSONEncoder.default = patched_default





import os
import json
from pathlib import Path

def view_for(view_name, fname, glob):
    view_name = view_name.replace('.', '_')
    if fname.endswith(('.csv', '.tsv')):
        return "CREATE VIEW \"{}\" AS SELECT * FROM read_csv_auto('{}', header=true)".format(view_name, glob)
    elif fname.endswith('.parquet'):
        return "CREATE VIEW \"{}\" AS SELECT * FROM '{}'".format(view_name, glob)
    elif fname.endswith(('.ndjson', '.jsonl')):
        return "CREATE VIEW \"{}\" AS SELECT * FROM read_ndjson_auto('{}')".format(view_name, glob)

def create_views(dirname):
    rv = []

    # Add in sorted order so the user sees alphabetically stable sort
    for f in sorted(os.scandir(dirname), key=lambda x: x.path):
        fname = f.path
        if f.is_dir():
            files = list(os.scandir(f.path))

            if not files:
                continue

            # We only sniff the first file, we assume all files in the directory
            # will have the same extension and shape. YOLO.
            file = files[0]
            rv.append(view_for(Path(fname).stem, file.path, os.path.join(fname, '*' + Path(file).suffix)))
        else:
            rv.append(view_for(Path(fname).stem, fname, fname))

    return [x for x in rv if x]



#  Copyright (c)  2020, salesforce.com, inc.
#  All rights reserved.
#  SPDX-License-Identifier: BSD-3-Clause
#  For full license text, see the COPYING file in the repo root or https://opensource.org/licenses/BSD-3-Clause

import threading


def debounce(wait_time):
    """
    Decorator that will debounce a function so that it is called after wait_time seconds
    If it is called multiple times, will wait for the last call to be debounced and run only this one.
    See the test_debounce.py file for examples
    """

    def decorator(function):
        def debounced(*args, **kwargs):
            def call_function():
                debounced._timer = None
                return function(*args, **kwargs)

            if debounced._timer is not None:
                debounced._timer.cancel()

            debounced._timer = threading.Timer(wait_time, call_function)
            debounced._timer.start()

        debounced._timer = None
        return debounced

    return decorator



import asyncio
import duckdb
from .debounce import debounce
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler, LoggingEventHandler
from datasette.database import Database, Results
from .ddl import create_views
from .winging_it import ProxyConnection

class SchemaEventHandler(FileSystemEventHandler):
    """React to files being added/removed from the watched directory."""

    def __init__(self, reload):
        super().__init__()

        self.reload = reload

    @debounce(1)
    def on_event(self):
        self.reload()

    def on_moved(self, event):
        super().on_moved(event)
        self.on_event()

    def on_created(self, event):
        super().on_created(event)
        self.on_event()

    def on_deleted(self, event):
        super().on_deleted(event)
        self.on_event()

    def on_modified(self, event):
        super().on_modified(event)
        self.on_event()

def create_directory_connection(directory):
    raw_conn = duckdb.connect()
    conn = ProxyConnection(raw_conn)

    for create_view_stmt in create_views(directory):
        conn.conn.execute(create_view_stmt)

    return conn

class DuckDatabase(Database):
    def __init__(self, ds, directory=None, file=None, httpfs=None, watch=None):
        super().__init__(ds)

        self.engine = 'duckdb'

        if directory:
            conn = create_directory_connection(directory)

            def reload():
                self.conn.conn.close()
                self.conn = create_directory_connection(directory)

            event_handler = SchemaEventHandler(reload)
            observer = Observer()
            observer.schedule(event_handler, directory, recursive=True)
            observer.start()
        elif file:
            raw_conn = duckdb.connect(file, read_only=True)
            conn = ProxyConnection(raw_conn)
        else:
            raise Exception('must specify directory or file')

        if httpfs:
            conn.conn.execute('install httpfs;').fetchall()
            conn.conn.execute('load httpfs;').fetchall()

        self.conn = conn

    @property
    def size(self):
        # TODO: implement this? Not sure if it's useful.
        return 0

    async def execute_fn(self, fn):
        if self.ds.executor is None:
            raise Exception('non-threaded mode not supported')

        def in_thread():
            return fn(self.conn)

        return await asyncio.get_event_loop().run_in_executor(
            self.ds.executor, in_thread
        )

    async def execute_write_fn(self, fn, block=True):
        if self.ds.executor is None:
            raise Exception('non-threaded mode not supported')

        def in_thread():
            return fn(self.conn)

        # We lie, we'll always block.
        return await asyncio.get_event_loop().run_in_executor(
            self.ds.executor, in_thread
        )




import re
import sqlglot
import sqlite3
table_xinfo_re = re.compile('^PRAGMA table_xinfo[(](.+)[)]')
table_info_square_re = re.compile('^PRAGMA table_info[(]\[(.+)][)]')

NO_OP_SQL = 'SELECT 0 WHERE 1 = 0'


def rewrite(sql):
    # print('rewrite: {}'.format(sql))

    if ' DATE(' in sql or ' date(' in sql:
        sql = NO_OP_SQL

    sql = sql.replace('<> ""', "<> ''")
    sql = sql.replace('!= ""', "!= ''")

    # This throws a duckdb exception if the column is not valid JSON,
    # but Datasette is only expecting SQLite exceptions.
    #
    # We could maybe translate that specific exception? Very sleazy.
    #
    # DuckDB also doesn't support json_each, though, so why bother?
    if ' json_type(' in sql or ' JSON_TYPE(' in sql:
        sql = NO_OP_SQL

    sql = sql.replace('"????-??-*"', "'????-??-*'")

    # TODO: once sqlglot #1066 lands, we can remove this
    if ' GLOB ' in sql or ' glob ' in sql:
        print('WARNING: rewriting a GLOB to a LIKE, this is almost certainly wrong')
        sql = sql.replace(' GLOB ', ' LIKE ').replace(' glob ', ' LIKE ')

    if sql == 'PRAGMA schema_version':
        sql = 'SELECT 0'

    if sql == 'select 1 from sqlite_master where tbl_name = "geometry_columns"':
        sql = NO_OP_SQL

    if sql == 'select name from sqlite_master where type="table"':
        sql = "select name from sqlite_master where type='table'"

    # Thwart https://github.com/simonw/datasette/blob/0b4a28691468b5c758df74fa1d72a823813c96bf/datasette/utils/__init__.py#L1120-L1127
    if sql.startswith('explain '):
        # This triggers a fallback path where it assumes the params based on a regex.
        # An alternative: use sqlglot to find params, then return a table shape
        # that matches what datasette is expecting.
        raise sqlite3.DatabaseError()

    # DuckDB doesn't support table_xinfo, so use table_info with a faked hidden column
    m = table_xinfo_re.search(sql)
    if m:
        sql = 'SELECT *, 0 FROM pragma_table_info({})'.format(m.group(1))

    m = table_info_square_re.search(sql)
    if m:
        sql = 'PRAGMA table_info("{}")'.format(m.group(1))

    # DuckDB doesn't support this pragma.
    # Luckily, Parquet doesn't have foreign keys, so just return no rows
    if sql.startswith('PRAGMA foreign_key_list'):
        sql = NO_OP_SQL

    # DuckDB doesn't support this pragma.
    # Luckily, Parquet doesn't have indexes, so just return no rows
    if sql.startswith('PRAGMA index_list'):
        sql = NO_OP_SQL

    if sql.startswith('PRAGMA recursive_triggers'):
        sql = NO_OP_SQL

    # This is some query to discover if FTS is enabled?
    if 'VIRTUAL TABLE%USING FTS' in sql:
        sql = NO_OP_SQL

    # Transpile queries, eg [test] is not a valid way to quote a table
    # in DuckDB.
    #print('before transpile: {}'.format(sql))
    if not sql.startswith('PRAGMA') and not sql.startswith('COPY ') and not "from '" in sql:
        sql = sqlglot.transpile(sql, read='sqlite', write='duckdb')[0]

    #print('after transpile: {}'.format(sql))

    return sql




import time
from .rewrite import rewrite, NO_OP_SQL

# A collection of classes to provide a facade that mimics the sqlite3 DB-API
# interface.

class Row:
    def __init__(self, columns, tpl):
        self.columns = columns
        self.tpl = tpl

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.tpl[key]
        else:
            return self.tpl[self.columns[key]]

def fixup_params(sql, parameters):
    if isinstance(parameters, dict):
        # On the custom SQL page, Datasette jams any query parameter it finds
        # into the parameters for the backend. DuckDB is strict on unexpected
        # parameters, so try to remove them

        new_parameters = {}
        for k, v in parameters.items():
            if ':{}'.format(k) in sql:
                new_parameters[k] = v

        parameters = new_parameters

    # Sometimes we skip queries that DuckDB can't handle, eg DATE(...) facet queries.
    # If the old query had parameters, sending them with the new query will
    # cause an assertion to fail. So return an empty list of parameters.
    if sql == NO_OP_SQL:
        return sql, []

    if isinstance(parameters, (tuple, list)):
        return sql, parameters
    else:
        new_params = []
        for i, (k, v) in enumerate((parameters or {}).items()):
            new_params.append(v)
            sql = sql.replace(':' + k, '${}'.format(i + 1))

        #print('new sql: {}'.format(sql))
        #print('new params: {}'.format(new_params))
        return sql, new_params

class ProxyCursor:
    def __init__(self, conn, existing_cursor=None):
        self.conn = conn

        if existing_cursor:
            self.cursor = existing_cursor
        else:
            self.cursor = self.conn.cursor()

    def execute(self, sql, parameters=None):
        #print('# params={} sql={}'.format(parameters, sql))
        sql = rewrite(sql)
        sql, parameters = fixup_params(sql, parameters)

        #print('## params={} sql={}'.format(parameters, sql))
        t = time.time()
        rv = self.cursor.execute(sql, parameters)
        #print('took {}'.format(time.time() - t))
        return rv

    def fetchone(self):
        tpl = self.cursor.fetchone()

        if not tpl:
            return tpl

        columns = {}
        for i, x in enumerate(self.cursor.description):
            columns[x[0]] = i

        return Row(columns, tpl)

    def fetchmany(self, size=1):
        tpls = self.cursor.fetchmany(size)
        columns = {}
        for i, x in enumerate(self.cursor.description):
            columns[x[0]] = i

        return [Row(columns, tpl) for tpl in tpls]

    def fetchall(self):
        tpls = self.cursor.fetchall()
        columns = {}
        for i, x in enumerate(self.cursor.description):
            columns[x[0]] = i

        return [Row(columns, tpl) for tpl in tpls]

    def __iter__(self):
        return self

    def __next__(self):
        rv = self.cursor.fetchone()

        if rv == None:
            raise StopIteration

        columns = {}
        for i, x in enumerate(self.cursor.description):
            columns[x[0]] = i
        rv = Row(columns, rv)
        return rv

    def __getattr__(self, name):
        return getattr(self.cursor, name)

class ProxyConnection:
    def __init__(self, conn):
        self.conn = conn

    def __enter__(self):
        pass

    def __exit__(self, exc_type,exc_value, exc_traceback):
        pass

    def execute(self, sql, parameters=None):
        #print('! params={} sql={}'.format(parameters, sql))
        sql = rewrite(sql)
        #print('! rewritten sql={}'.format(sql))
        sql, parameters = fixup_params(sql, parameters)
        #print('!! params={} sql={}'.format(parameters, sql))
        rv = self.conn.execute(sql, parameters)

        return ProxyCursor(self.conn, rv)

    def fetchall(self):
        raise Exception('TODO: ProxyConnection.fetchall is not implemented')

    def set_progress_handler(self, handler, n):
        pass

    def cursor(self):
        return ProxyCursor(self.conn)





from datasette import hookimpl

PLUGIN_NAME = 'datasette-parquet'

@hookimpl
def startup(datasette):
    config = datasette.plugin_config(
        PLUGIN_NAME
    )

    if not config:
        return

    from .ducky import DuckDatabase
    from .patches import monkey_patch

    monkey_patch()

    for db_name, options in config.items():
        if not 'directory' in options and not 'file' in options:
            raise Exception('datasette-parquet: expected directory or file key for db {}'.format(db))

        if 'directory' in options:
            directory = options['directory']
            db = DuckDatabase(datasette, directory=directory, watch=options.get('watch', False) == True)
            datasette.add_database(db, db_name)
        else:
            file = options['file']
            db = DuckDatabase(datasette, file=file)
            datasette.add_database(db, db_name)

